/**
 * @file main.cpp
 * @brief main program file for the scale space meshing method
 * @author Julie Digne julie.digne@liris.cnrs.fr
 * @date 2012/11/14
 * @copyright This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#define _USE_MATH_DEFINES
#include <cmath>

#include <iostream>
#include <sstream>
#include <ctime>
#include <vector>

#include "src/Octree.h"
#include "src/ScaleSpace.h"
#include "src/utilities.h"
#include "src/Vertex.h"
#include "src/Mesher.h"
#include "src/FileIO.h"

#include "src/types.h"
#include "src/cmdLine.h"

using namespace std;

/** @brief display the usage of the program
 */
void usage()
{
    std::cout<<"USAGE"<<std::endl;
    std::cout<<std::endl;
    std::cout<<std::endl;
    std::cout<<"scalespacemeshing INPUT OUTPUT"
             <<" -r <radius> -N <Niter> -p"<<std::endl;
    std::cout<<std::endl;
    std::cout<<"INPUT  (mandatory) input_file.txt is a ascii file containing the oriented"<<std::endl;
    std::cout<<"       points."<<std::endl;
    std::cout<<std::endl;
    std::cout<<"OUTPUT (mandatory) output_file. The output file will be named"<<std::endl;
    std::cout<<"       \"$OUTPUT_fine_scale.ply\" and will contain the output reconstructed"<<std::endl;
    std::cout<<"       surface. If (-v) option is set the program will also output"<<std::endl;
    std::cout<<"       \"$OUTPUT_coarse_scale.ply\" (see -v option)"<<std::endl;
    std::cout<<"       The format is the Stanford PLY format (ascii)."<<std::endl;
    std::cout<<std::endl;
    std::cout<<"-r     (recommended) Radius that will be used to reconstruct"<<std::endl;
    std::cout<<"       the surface. If not set a coarse estimation will be used."<<std::endl;
    std::cout<<std::endl;
    std::cout<<"-N     (recommended) Number of scale space iterations. If not set or if a negative number is"<<std::endl;
    std::cout<<"       set the default number of iterations will be used (Niter = 4)."<<std::endl;
    std::cout<<std::endl;
    std::cout<<"-v     (option) verbose tag that will cause the program to output an additional"<<std::endl;
    std::cout<<"       mesh: the coarse scale mesh before back projection that will be named"<<std::endl;
    std::cout<<"       \"$OUTPUT_coarse_scale.ply\"."<<std::endl;
    std::cout<<std::endl;
    std::cout<<"-p     (recommended) will perform the computation in parallel."<<std::endl;
  
  
}

/**
 * @brief main function for the ball pivoting reconstruction
 * @param argc
 * @param argv
 * @return 1 if the program exited successfully
 */
int main(int argc, char **argv)
{
    CmdLine cmd;
    //handling command line options
    std::stringstream f;
    string infile, outfile;
    double radius = -1;
    int parallel_flag = -1;
    int verbose_flag = -1;
    int niter = 4;

    cmd.add(make_option('r', radius));
    cmd.add(make_option('N', niter));
    cmd.add(make_switch('p'));
    cmd.add(make_switch('v'));


    try {
        cmd.process(argc, argv);
    } catch(std::string str) {
        std::cerr << "Error: " << str << std::endl<<std::endl;
        usage();
        return EXIT_FAILURE;
    }

    if(argc < 3)
    {
        usage();
        return EXIT_FAILURE;
    }
    
    infile = argv[1];
    outfile = argv[2];

    std::cout<<"Input file :"<<argv[1]<<std::endl;
    std::cout<<"Output file: "<<argv[2]<<std::endl;

    if(cmd.used('p')) parallel_flag = 1;
    if(cmd.used('v')) verbose_flag = 1;

    if(! cmd.used('r') || (radius < 0))
    {
      std::cout<<"No radius or negative radius given!"<<std::endl;
      std::cout<<"Estimating a radius ... "<<std::flush;
      radius = FileIO::estimateRadius(infile.c_str());
      std::cout<<radius<<std::endl;;
    }
    if( (! cmd.used('N')) || (niter < 0) )
    {
        niter = 4;
        std::cout<<"No number of iterations given or negative number of";
        std::cout<<" iterations, using default: "<<niter<<std::endl;
    }

    time_t start,end, glob_start,glob_end;

    Octree octree;

    std::time(&start);

    bool ok;
    ok = FileIO::readAndSortPoints(infile.c_str(), octree, radius); 

    if( !ok )
    {
        std::cerr<<"Pb opening the file; exiting."<<std::endl;
        return EXIT_FAILURE;
    }
    std::time(&end);

    std::cout<<"Parameters: radius: "<<radius<<" ; number of iterations "<<niter<<"."<<std::endl;
    std::cout<<"Octree with depth "<<octree.getDepth()<<" created."<<std::endl;
    std::cout<<"Octree contains "<<octree.getNpoints()<<" points."<<std::endl;
    std::cout<<"The bounding box size is "<<octree.getSize()<<std::endl;
    std::cout<<"Reading and sorting the points took "<<difftime(end,start)
             <<" s."<<std::endl;

    std::cout<<"Octree statistics"<<std::endl;
    octree.printOctreeStat();

    if(radius < 0) 
    {
        radius = 0.5 * octree.getSmallestCellSize();
    }

    //creating the scale space structure
    std::time(&start);
    std::time(&glob_start);
    ScaleSpace scalespace(&octree, 2.0*radius, niter, 0);

    if(parallel_flag == 1)
        scalespace.parallelApplyScaleSpace();
    else
        scalespace.applyScaleSpace();

    std::time(&end);
    std::cout<<"Applying "<<niter<<" scale space iterations took "
             <<difftime(end,start) <<" s."<<std::endl;

    Mesher mesher(&octree, radius, scalespace.getSetIndex());

    std::time(&start);
    if(parallel_flag == 1)
        mesher.parallelReconstruct();
    else
        mesher.reconstruct();
    std::time(&end);

    std::cout<<"Reconstructed mesh: "<<mesher.nVertices()
             <<" vertices; "<<mesher.nFacets()
             <<" facets. "<<std::endl;
    std::cout<<mesher.nBorderEdges()<<" border edges"<<std::endl;
    std::cout<<"Reconstructing the mesh took "<<difftime(end,start)
             <<" s."<<std::endl;

    std::cout<<"Filling the holes... "<<std::flush;
    std::time(&start);
    mesher.fillHoles();
    std::time(&end);
    std::time(&glob_end);
    std::cout<<difftime(end,start)<<" s."<<std::endl;
    std::cout<<"Final mesh: "<<mesher.nVertices()<<" vertices; "
             <<mesher.nFacets()<<" facets. "<<std::endl;
    std::cout<<mesher.nBorderEdges()<<" border edges"<<std::endl;

    std::cout<<"Total computation time "<<difftime(glob_end, glob_start)
             <<" s."<<std::endl;

    if(verbose_flag == 1)
    {
        std::string outputc = outfile;
        outputc.append("_coarse_scale.ply");
        std::cout<<"saving the coarse mesh in "<<outputc<<std::endl;
        if(! FileIO::saveMesh(outputc.c_str(), mesher))
        {
            std::cerr<<"Pb saving the mesh "<<outputc.c_str()
            <<" ; exiting."<<std::endl;
            return EXIT_FAILURE;
        }
    }

    std::string outputf = outfile;
    outputf.append("_fine_scale.ply");
    std::cout<<"saving the back-projected mesh in "<<outputf<<std::endl;
    if(! FileIO::saveBackProjectedMesh(outputf.c_str(), mesher))
    {
        std::cerr<<"Pb saving the back projected mesh "<<outputf.c_str()
                 <<" ; exiting."<<std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
