/** @file FileIO.cpp
 * @brief file defining methods to read points from a file
 * @author Julie Digne julie.digne@liris.cnrs.fr
 * @date 2012/10/22
 * @copyright This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "FileIO.h"
#include<iostream>
#include<fstream>
#include<istream>
#include<ostream>
#include<sstream>
#include<algorithm>
#include<deque>
#include<limits>

using namespace std;

FileIO::FileIO()
{
}

FileIO::~FileIO()
{
}

double FileIO::estimateRadius(const char* filename)
{
    ifstream in;
    in.open(filename);

    if(!in)
    {
        std::cerr<<"File "<<filename<<" could not be opened"<<std::endl;
        return false;
    }

    string line;

    getline(in, line);

    double x,y,z;
    in >> x >> y >> z;

    double xmin, ymin, zmin, xmax, ymax, zmax;
    xmin = xmax = x;
    ymin = ymax = y;
    zmin = zmax = z;

    unsigned int npts = 1;
    while(getline(in,line))
    {
        in >> x >> y >> z;
        xmin = std::min(x,xmin);
        xmax = std::max(x,xmax);
        ymin = std::min(y,ymin);
        ymax = std::max(y,ymax);
        zmin = std::min(z,zmin);
        zmax = std::max(z,zmax);
        ++npts;
    }
    in.close();

    double lx = xmax - xmin;
    double ly = ymax - ymin;
    double lz = zmax - zmin;

    double size = std::max(lx,ly);
    size = std::max(lz,size);

    return std::sqrt(20.0/(double)npts)*size;
}


bool FileIO::readAndSortPoints(const char* filename, Octree& octree, double 
min_radius)
{
    ifstream in;
    in.open(filename);

    if(!in)
    {
        std::cerr<<"File "<<filename<<" could not be opened"<<std::endl;
        return false;
    }

    string firstline;
    getline(in, firstline);

    istringstream line_in(firstline);
    string word;
    int nword = 0;
    while (line_in>> word)
        nword++;

    if(nword == 3)
    {
        cerr<< "Only three doubles per line: unoriented points? "
            << "This program needs oriented points!"<<endl;
        return false;
    }

    if( nword != 6)
    {
        cerr<<"each point must be given by 6 values (position + normal) : "
        <<"x y z nx ny nz"<<endl;
        return false;
    }

    in.clear() ;
    in.seekg(0, ios::beg);//starting back from the beginning

    double x,y,z,nx,ny,nz;
    in >> x >> y >> z >> nx >> ny >> nz;

    deque<Vertex> input_vertices;
    input_vertices.push_back(Vertex(x,y,z,nx,ny,nz));

    double xmin, ymin, zmin, xmax, ymax, zmax;
    xmin = xmax = x;
    ymin = ymax = y;
    zmin = zmax = z;

    while( in >> x >> y >> z >> nx >> ny >> nz)
    {
        input_vertices.push_back(Vertex(x,y,z,nx,ny,nz));
        xmin = std::min(x,xmin);
        xmax = std::max(x,xmax);
        ymin = std::min(y,ymin);
        ymax = std::max(y,ymax);
        zmin = std::min(z,zmin);
        zmax = std::max(z,zmax);
    }
    in.close();

    std::cout<<input_vertices.size()<<" points read"<<std::endl;

    double lx = xmax - xmin;
    double ly = ymax - ymin;
    double lz = zmax - zmin;

    double size = std::max(lx,ly);
    size = std::max(lz,size);
    size = 1.1 * size;//loose bounding box around the object

    double margin;
    if(min_radius > 0)
    {
        unsigned int depth = (unsigned int)ceil( log2( size / (min_radius) ));
        //adapting the bouding box size so that the smallest cell size is 
        //exactly 2*min_radius
        double adapted_size = pow2(depth) * min_radius;
        //margins of the bouding box around the object
        margin = 0.5 * (adapted_size - size); 
        size = adapted_size;
        octree.setDepth(depth);
    }
    else
    {
        margin = 0.05 * size; //margins of the bouding box around the object
    }

    double ox = xmin - margin;
    double oy = ymin - margin;
    double oz = zmin - margin;
    Point origin(ox,oy,oz);

    octree.initialize(origin, size);

    //add the points to the set with index 0 (initial set)
    octree.addInitialPoints(input_vertices.begin(), input_vertices.end());

    return true;
}

bool FileIO::savePoints(const char* filename, Octree& octree,
                        unsigned int index)
{ 
    ofstream out;
    out.open(filename);
    out.precision( numeric_limits<double>::digits10 + 1);

    if(!out)
        return false;

    OctreeNode *node = octree.getRoot();
    saveContent(node, index, out);

    out.close();

    return true;
}




bool FileIO::saveMesh(const char* filename, Mesher& mesher)
{ 
    ofstream out;
    out.open(filename);
    out.precision( numeric_limits<double>::digits10 + 1);
    
    if(!out)// file couldn't be opened
    { 
        cerr << "Error: file could not be opened" << endl;
        return false;
    }

    out << "ply" << endl;
    out << "format ascii 1.0" << endl;
    out << "comment Coarse Scale Mesh generated by the Scale Space Meshing"
        << " Algorithm" << endl;
    out << "element vertex "<< mesher.nVertices() << endl;
    out << "property float x" << endl;
    out << "property float y" << endl;
    out << "property float z" << endl;
    out << "property float nx" << endl;
    out << "property float ny" << endl;
    out << "property float nz" << endl;
    out << "element face "<< mesher.nFacets() << endl;
    out << "property list uchar int vertex_indices"<< endl;
    out << "end_header" << endl;

    for( Vertex_star_deque::const_iterator vi = mesher.vertices_begin();
        vi != mesher.vertices_end(); ++vi)
    {
        Vertex *v = *vi;
        out<< *v << endl;
    }

    bool success = saveOrientedFacets(mesher, out);
    return success;
}

bool FileIO::saveBackProjectedMesh(const char* filename, Mesher& mesher)
{ 
    ofstream out;
    out.open(filename);
    
    out.precision( numeric_limits<double>::digits10 + 1);
    if(!out)
    { // file couldn't be opened
        cerr << "Error: file could not be opened" << endl;
        return false;
    }

    out << "ply" << endl;
    out << "format ascii 1.0" << endl;
    out << "comment Back projected mesh generated by the Scale Space Meshing"
        << " Algorithm" << endl;
    out << "element vertex "<< mesher.nVertices() << endl;
    out << "property float x" << endl;
    out << "property float y" << endl;
    out << "property float z" << endl;
    out << "property float nx" << endl;
    out << "property float ny" << endl;
    out << "property float nz" << endl;
    out << "element face "<< mesher.nFacets() << endl;
    out << "property list uchar int vertex_indices"<< endl;
    out << "end_header" << endl;

    for( Vertex_star_deque::const_iterator vi = mesher.vertices_begin();
        vi != mesher.vertices_end(); ++vi)
    {
        const Vertex *v = (*vi)->getOriginal();
        out<< *v << endl;
    }

    bool success = saveOrientedFacets(mesher, out);
    return success;
}

void FileIO::saveContent(OctreeNode* node, unsigned int index, ofstream& f)
{
    if(node->getDepth() != 0)
    {
        for(int i = 0; i < 8 ;i++)
            if(node->getChild(i) != NULL)
                saveContent(node->getChild(i), index, f);
    }
    else if(node->getNpts(index) != 0)
    {
        Vertex_deque::const_iterator iter;
        for(iter = node->points_begin(index);
            iter != node->points_end(index);++iter)
        {
            f << *iter;
        }
    }
}


bool FileIO::saveOrientedFacets(Mesher &mesher,ofstream& out)
{
    for( Facet_star_deque::const_iterator fi = mesher.facets_begin();
        fi != mesher.facets_end(); ++fi)
    {
        out << 3 << "\t";
        const Facet* f = *fi;

        Vertex *v0 = f->vertex(0);
        Vertex *v1 = f->vertex(1);
        Vertex *v2 = f->vertex(2);

        double nx = v0->nx() + v1->nx() + v2->nx();
        double ny = v0->ny() + v1->ny() + v2->ny();
        double nz = v0->nz() + v1->nz() + v2->nz();

        double tx = v1->x() - v0->x();
        double ty = v1->y() - v0->y();
        double tz = v1->z() - v0->z();

        double sx = v2->x() - v0->x();
        double sy = v2->y() - v0->y();
        double sz = v2->z() - v0->z();

        double cprodx, cprody, cprodz;
        cross_product(tx, ty, tz, sx, sy, sz, cprodx, cprody, cprodz);

        //saving the vertex indices in clockwise order around the oriented 
        //normal
        if( nx * cprodx + ny * cprody + nz * cprodz > 0)
        {
            out << v0->index() <<"\t";
            out << v1->index() <<"\t";
            out << v2->index() <<endl;
        }
        else
        {
            out << v0->index() <<"\t";
            out << v2->index() <<"\t";
            out << v1->index() <<endl;
        }
    }
    return true;
}


